#!/usr/bin/env python
#     Copyright 2015, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Softwar where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

b = range(10000)

def getGeneratorFunction():
   def f():
      for i in b:
         yield i

   return f

def getGeneratorExpression():
   return ( i for i in b )


import time

start = time.time()

f = getGeneratorFunction()

for x in range(1000):
   r = list(f())

end = time.time()

func_time = end - start

start = time.time()

for x in range(1000):
   r = list(getGeneratorExpression())

end = time.time()

genexpr_time = end - start

print "Generator Function took", func_time
print "Generator Expression took", genexpr_time
